'''
[Pix2Pix]
Isola, Phillip, et al. Image-to-image translation with conditional adversarial networks. CVPR 2017.
Code reference: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix

[运行]
python Practice_19_pix2pix.py

[环境要求]
- Pytorch
- torchvision
- opencv
- matplotlib
- pillow
- tqdm

[数据]
在本文件夹的data/paired目录下下载以下数据用作训练
wget http://efrosgans.eecs.berkeley.edu/pix2pix/datasets/facades.tar.gz
tar -xf facades.tar.gz
rm -f facades.tar.gz
然后将文件结构整理如下
data/paired/facades
├── test
│   ├── 1.jpg
│   ├── ...
│   └── 106.jpg
├── data/train
│   ├── 1.jpg
│   ├── ...
│   └── 400.jpg
└── val
    ├── 1.jpg
    ├── ...
    └── 100.jpg
'''

import os
from PIL import Image
import matplotlib.pyplot as plt

import torchvision.transforms as transforms
import torchvision.models as models

import numpy as np
from datetime import datetime
import time
import logging
import random
import cv2
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.autograd as autograd
from torch.autograd import Variable
from torch.nn import Linear, Conv2d, BatchNorm2d, LayerNorm, LeakyReLU, ConvTranspose2d, ReLU, Tanh, InstanceNorm2d
from torch.nn import ReflectionPad2d, ReplicationPad2d

from torch.utils.data import DataLoader
import torchvision.datasets as dset
import torchvision
import itertools
from tqdm import tqdm


'''辅助函数''' 
# 读图像
def load_image(filename):
    transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.5, 0.5, 0.5],std=[0.5,0.5,0.5]),
    ])
    
    img = Image.open(filename)
    img = transform(img)
    return img.unsqueeze(dim=0)   

# 保存图像
def save_image(img, filename):
    tmp = ((img.detach().numpy().transpose(1, 2, 0) + 1.0) * 127.5).astype(np.uint8)
    cv2.imwrite(filename, cv2.cvtColor(tmp, cv2.COLOR_RGB2BGR))

# 网络参数初始化
def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
        if not(m.bias is None):
            m.bias.data.fill_(0)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)
        
        
# 定义计算对抗损失的GANLoss类
class GANLoss(nn.Module):
    def __init__(self, LSGAN=True):
        super(GANLoss, self).__init__()
        # 定义两个目标张量，分别表示True和False的1.0和0.0
        self.register_buffer('real_label', torch.tensor(1.0))
        self.register_buffer('fake_label', torch.tensor(0.0))
        if LSGAN: 
            # LSGAN loss
            self.loss = nn.MSELoss()
        else:
            # 原版loss
            self.loss = nn.BCEWithLogitsLoss()

    def forward(self, prediction, target_is_real):
        if target_is_real:
            target_tensor = self.real_label
        else:
            target_tensor = self.fake_label
        # 将目标张量扩展到与输出的特征的维度一致
        target_tensor = target_tensor.expand_as(prediction)
        return self.loss(prediction, target_tensor)

    
'''网络定义'''
# U-Net包含编码器和解码器，编码器和解码器分别包含N=8个下采的卷积层和上采的卷积层，共16个卷积层。第i层和第17-i层跨层连接
class UnetGenerator(nn.Module):
    def __init__(self, num_downs=8, ngf=64):
        super(UnetGenerator, self).__init__()
        # 构造最内侧的2层
        unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, None, innermost=True)
        # 构成次内层的6层，它们的卷积核数目都是512
        for i in range(num_downs - 5):
            unet_block = UnetSkipConnectionBlock(ngf * 8, ngf * 8, unet_block)
        # 每次构造外侧的2层，且卷积核数目减半
        unet_block = UnetSkipConnectionBlock(ngf * 4, ngf * 8, unet_block)
        unet_block = UnetSkipConnectionBlock(ngf * 2, ngf * 4, unet_block)
        unet_block = UnetSkipConnectionBlock(ngf, ngf * 2, unet_block)
        # 构造最外侧的2层，输入3通道的图像，输出3通道的图像
        self.model = UnetSkipConnectionBlock(3, ngf, unet_block, outermost=True)

    def forward(self, input):
        return self.model(input)

# UnetSkipConnectionBlock类定义了包含跨层连接的U-Net子模块
# 形式为下采层→内侧的submodule→上采层，其中该模块的输入直接串联到输出
# submodule为已经构造好的更内侧的网络部分
# 通道数变化为：input_nc→inner_nc→outer_nc→input_nc+outer_nc
# outermost和innermost分别指示当前部分是否为网络的最外侧和最内侧
class UnetSkipConnectionBlock(nn.Module):
    def __init__(self, outer_nc, inner_nc, submodule=None, outermost=False, innermost=False):
        super(UnetSkipConnectionBlock, self).__init__()
        self.outermost = outermost
        input_nc = outer_nc
        downconv = nn.Conv2d(input_nc, inner_nc, kernel_size=4,
                             stride=2, padding=1, bias=False)
        downrelu = nn.LeakyReLU(0.2, True)
        downnorm = nn.BatchNorm2d(inner_nc)
        uprelu = nn.ReLU(True)
        upnorm = nn.BatchNorm2d(outer_nc)

        if outermost:
            # 最外侧为两层卷积层，夹在已定义好的submodule外侧
            upconv = nn.ConvTranspose2d(inner_nc * 2, outer_nc,
                            kernel_size=4, stride=2, padding=1)
            down = [downconv]
            up = [uprelu, upconv, nn.Tanh()]
            model = down + [submodule] + up
        elif innermost:
            # 最内侧为两层卷积层，不含更内侧的submodule
            upconv = nn.ConvTranspose2d(inner_nc, outer_nc,
                            kernel_size=4, stride=2, padding=1, bias= False)
            down = [downrelu, downconv]
            up = [uprelu, upconv, upnorm]
            model = down + up
        else:
            # 中间与最外侧类似，包含两层卷积层，夹在submodule外侧
            # 与最外侧不同，中间的卷积层遵循卷积+批标准化+激活层的形式
            # 而最外侧的卷积层不含批标准化层，最后一层激活层改为tanh
            upconv = nn.ConvTranspose2d(inner_nc * 2, outer_nc,
                            kernel_size=4, stride=2, padding=1, bias= False)
            down = [downrelu, downconv, downnorm]
            up = [uprelu, upconv, upnorm]
            model = down + [submodule] + up

        self.model = nn.Sequential(*model)

    def forward(self, x):
        if self.outermost:
            return self.model(x)
        else:   # 内侧的层之间需要添加跨层连接
            return torch.cat([x, self.model(x)], 1)
        
print(UnetGenerator(num_downs=8, ngf=64))

class NLayerDiscriminator(nn.Module):
    def __init__(self, ndf=64, n_layers=3):
        super(NLayerDiscriminator, self).__init__()
        sequence = [
            nn.Conv2d(6, ndf, kernel_size=4, stride=2, padding=1),
            nn.LeakyReLU(0.2, True)
        ]
        nf_mult = 1
        nf_mult_prev = 1
        # 随着层数的加深逐渐增加卷积核的数量
        for n in range(1, n_layers):  
            nf_mult_prev = nf_mult
            nf_mult = min(2 ** n, 8)
            sequence += [
                nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult, kernel_size=4, stride=2, padding=1, bias=False),
                nn.BatchNorm2d(ndf * nf_mult),
                nn.LeakyReLU(0.2, True)
            ]

        nf_mult_prev = nf_mult
        nf_mult = min(2 ** n_layers, 8)
        sequence += [
            nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult, kernel_size=4, stride=1, padding=1, bias= False),
            nn.BatchNorm2d (ndf * nf_mult),
            nn.LeakyReLU(0.2, True)
        ]

        # 输出层的通道数降为1，空间坐标上的每个元素指示真假
        sequence += [nn.Conv2d(ndf * nf_mult, 1, kernel_size=4, stride=1, padding=1)]  
        self.model = nn.Sequential(*sequence)

    def forward(self, input):
        return self.model(input)

print(NLayerDiscriminator(n_layers=3, ndf=64))    


# 定义pix2pix网络模型及训练框架
class Pix2PixModel(nn.Module):
    def __init__(self):
        super(Pix2PixModel, self).__init__()
        # 定义G网络为U-Net，对于256*256的图像，编码器下采8次
        # 定义D网络，包含3+2层卷积层
        self.netG = UnetGenerator(num_downs=8, ngf=64)
        self.netD = NLayerDiscriminator(n_layers=3, ndf=64)

        # 定义对抗损失和L1重建损失
        self.criterionGAN = GANLoss(LSGAN=False)
        self.criterionL1 = torch.nn.L1Loss()
        # 定义G和D的优化器
        self.optimizer_G = torch.optim.Adam(self.netG.parameters(), lr=0.0002, betas=(0.5, 0.999))
        self.optimizer_D = torch.optim.Adam(self.netD.parameters(), lr=0.0002, betas=(0.5, 0.999))

    # G网络将A映射为B
    def forward(self, real_A):
        return self.netG(real_A)

    # 计算D网络的梯度
    def backward_D(self, real_A, real_B, fake_B):
        # 对于生成图片，D网络需要判断其为False
        fake_AB = torch.cat((real_A, fake_B), 1)  
        pred_fake = self.netD(fake_AB)
        loss_D_fake = self.criterionGAN(pred_fake, False) # 公式(6.2)第二项
        # 对于真实图片，D网络需要判断其为True
        real_AB = torch.cat((real_A, real_B), 1)
        pred_real = self.netD(real_AB)
        loss_D_real = self.criterionGAN(pred_real, True) # 公式(6.2)第一项
        loss_D = (loss_D_fake + loss_D_real) * 0.5
        loss_D.backward()  # 计算梯度并反向传播
        return loss_D.data.mean().item()

    # 计算G网络的梯度
    def backward_G(self, real_A, real_B, fake_B):
        # G网络需要迷惑D网络判断为True
        fake_AB = torch.cat((real_A, fake_B), 1)
        pred_fake = self.netD(fake_AB)
        loss_G_GAN = self.criterionGAN(pred_fake, True) # 公式(6.2)第二项
        # 此外G网络还需要逼近真实的图像real_B
        loss_G_L1 = self.criterionL1(fake_B, real_B) * 100 # 公式(6.3)
        loss_G = loss_G_GAN + loss_G_L1
        loss_G.backward()  # 计算梯度并反向传播
        return loss_G_GAN.data.mean().item(), loss_G_L1.data.mean().item()
    
    # 训练流程，real_A和real_B为输入图像，网络学习real_A到real_B的映射
    def optimize_parameters(self, real_A, real_B):
        fake_B = self.forward(real_A)
        # 更新D网络
        self.optimizer_D.zero_grad() # 将D网络参数梯度初始化为0 
        loss_D = self.backward_D(real_A, real_B, fake_B.detach())       
        self.optimizer_D.step()    # 更新D网络参数        
        # 更新G网络
        self.optimizer_G.zero_grad() # 将G网络参数梯度初始化为0
        loss_G_GAN, loss_G_L1 = self.backward_G(real_A, real_B, fake_B)          
        self.optimizer_G.step()   # 更新G网络参数
        return fake_B, loss_D, loss_G_GAN, loss_G_L1


'''数据准备''' 
dataset = dset.ImageFolder(root='./data/paired/facades/data/',
                           transform=transforms.Compose([
                               transforms.Resize(288),
                               transforms.ToTensor(),
                               transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
                           ]))

dataloader = DataLoader(dataset, batch_size=1, shuffle=True, num_workers=2, drop_last=True)
print('Dataset size:', len(dataloader.dataset.imgs))


'''构建和初始化模型'''
model = Pix2PixModel()
if torch.cuda.is_available():
    model.cuda()
model.netG.apply(weights_init)
model.netD.apply(weights_init)
print('Model created.') 


'''网络训练'''
pbar = tqdm(range(50000), initial=0, smoothing=0.01, dynamic_ncols=True)

model.train()
iterator = iter(dataloader)
for idx in pbar:
    try:
        data, _ = next(iterator)
    except StopIteration:
        iterator = iter(dataloader)
        data, _ = next(iterator) 
        
    real_B = data[:,:,:,0:288].cuda()
    real_A = data[:,:,:,288:288*2].cuda()
    
    w = random.randint(0,288-256)
    h = random.randint(0,288-256)
    real_B = real_B[:,:,w:w+256,h:h+256]
    real_A = real_A[:,:,w:w+256,h:h+256]
    
    if np.random.rand(1) > 0.5:
        real_A = real_A[:,:,:,torch.arange(real_A.size(3) - 1, -1, -1).long()]
        real_B = real_B[:,:,:,torch.arange(real_B.size(3) - 1, -1, -1).long()]
    
    fake_B, loss_D, loss_G_GAN, loss_G_L1 = model.optimize_parameters(real_A, real_B)

    pbar.set_description(
        (
            f"d: {loss_D:.3f}; g: {loss_G_GAN:.3f}; l1: {loss_G_L1:.3f};"
        )
    )    

# 保存模型    
torch.save(model.state_dict(),  f"pix2pix-%05d.pt"%(idx))


'''测试模型'''
model.eval()
I = load_image('./data/paired/facades/test/1.jpg')
with torch.no_grad():
    real_B = I[:,:,:,0:256]
    real_A = I[:,:,:,256:256*2]   
    output = model(real_A.cuda()).cpu()
save_image(output[0], '1_pix2pix.jpg')